@@ -14,7 +14,6 @@ import (
 	"flag"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"math"
 	"os"
 	"os/exec"
@@ -41,6 +40,12 @@ var (
 	defaultCfg0 *Config
 	re          *regexp.Regexp
 	totalRam    = memory.TotalMemory()
+	makeJ       = fmt.Sprintf("-j%d", runtime.GOMAXPROCS(-1))
+	_           = stack
+	goos        = runtime.GOOS
+	goarch      = runtime.GOARCH
+	target      = fmt.Sprintf("%s/%s", goos, goarch)
+	nogcc       = goos == "windows" && goarch == "arm64" // We have no 32b mingw-gcc binary targeting windows/amd64 bit yet.
 
 	oTrace = flag.Bool("trc", false, "Print tested paths.")
 )
@@ -50,7 +55,45 @@ func init() {
 	isTesting = true
 	extendedErrors = true
 	var err error
-	if defaultCfg0, err = NewConfig(runtime.GOOS, runtime.GOARCH, []string{LongDouble64Flag(runtime.GOOS, runtime.GOARCH)}...); err != nil {
+	flags := []string{LongDouble64Flag(runtime.GOOS, runtime.GOARCH)}
+	if goos == "windows" && (goarch == "386" || goarch == "amd64") {
+		flags = append(flags,
+			"-mno-3dnow",
+			"-mno-abm",
+			"-mno-aes",
+			"-mno-avx",
+			"-mno-avx2",
+			"-mno-avx512cd",
+			"-mno-avx512er",
+			"-mno-avx512f",
+			"-mno-avx512pf",
+			"-mno-bmi",
+			"-mno-bmi2",
+			"-mno-f16c",
+			"-mno-fma",
+			"-mno-fma4",
+			"-mno-fsgsbase",
+			"-mno-lwp",
+			"-mno-lzcnt",
+			"-mno-mmx",
+			"-mno-pclmul",
+			"-mno-popcnt",
+			"-mno-prefetchwt1",
+			"-mno-rdrnd",
+			"-mno-sha",
+			"-mno-sse",
+			"-mno-sse2",
+			"-mno-sse3",
+			"-mno-sse4",
+			"-mno-sse4.1",
+			"-mno-sse4.2",
+			"-mno-sse4a",
+			"-mno-ssse3",
+			"-mno-tbm",
+			"-mno-xop",
+		)
+	}
+	if defaultCfg0, err = NewConfig(runtime.GOOS, runtime.GOARCH, flags...); err != nil {
 		panic(errorf("NewConfig: %v", err))
 	}
 
@@ -94,7 +137,6 @@ func exampleAST(rule int, src string) (r interface{}) {
 	defer func() {
 		if err := recover(); err != nil {
 			r = fmt.Sprintf("%v (%v:)", err, origin(5))
-			trc("%v\n%s", r, debug.Stack())
 		}
 	}()
 
@@ -230,7 +272,7 @@ type parallel struct {
 }
 
 func newParallel() *parallel {
-	limit := runtime.GOMAXPROCS(0)
+	limit := runtime.GOMAXPROCS(-1)
 	switch runtime.GOARCH {
 	case "386", "arm": // 32 bit targets
 		limit = 1
@@ -245,8 +287,8 @@ func newParallel() *parallel {
 }
 
 func (p *parallel) exec(run func()) {
-	p.limit <- struct{}{}
 	p.wg.Add(1)
+	p.limit <- struct{}{}
 
 	go func() {
 		defer func() {
@@ -328,7 +370,7 @@ func getCorpusFile(path string) ([]byte, error) {
 		return nil, err
 	}
 
-	return ioutil.ReadAll(f)
+	return io.ReadAll(f)
 }
 
 func h(v interface{}) string {
@@ -707,7 +749,7 @@ func TestTranslationPhase4(t *testing.T) {
 		"11.c": {},
 	}
 	switch fmt.Sprintf("%s/%s", runtime.GOOS, runtime.GOARCH) {
-	case "linux/s390x", "linux/loong64":
+	case "linux/s390x", "linux/loong64", "linux/mips64le":
 		blacklistCompCert["aes.c"] = struct{}{} // Unsupported endianness.
 	}
 	var files, ok, skip, fails int32
@@ -1072,6 +1114,8 @@ func testParserBug(t *testing.T, dir string, blacklist map[string]struct{}) {
 	t.Logf("files %v, skip %v, ok %v, fails %v", files, skip, ok, len(fails))
 }
 
+// ~/src/modernc.org/ccorpus2/
+
 func TestParse(t *testing.T) {
 	cfg := defaultCfg()
 	cfg.SysIncludePaths = append(cfg.SysIncludePaths, "Include") // benchmarksgame
@@ -1081,9 +1125,16 @@ func TestParse(t *testing.T) {
 		// Assertions are deprecated, not supported.
 		"950919-1.c": {},
 	}
+	blacklistDebian := map[string]struct{}{
+		"nbody-4.c": {}, // _Float16 not supported
+		"nbody-8.c": {}, // _Float16 not supported
+	}
 	switch fmt.Sprintf("%s/%s", runtime.GOOS, runtime.GOARCH) {
-	case "linux/s390x", "linux/loong64":
+	case "linux/s390x", "linux/loong64", "linux/mips64le":
 		blacklistCompCert["aes.c"] = struct{}{} // Unsupported endianness.
+	case "freebsd/amd64":
+		blacklistDebian["mandelbrot-6.c"] = struct{}{}
+		blacklistDebian["spectral-norm-5.c"] = struct{}{}
 	}
 	var files, ok, skip, fails int32
 	for _, v := range []struct {
@@ -1100,7 +1151,7 @@ func TestParse(t *testing.T) {
 		{cfg, "github.com/vnmakarov", nil},
 		{cfg, "sqlite-amalgamation", nil},
 		{cfg, "tcc-0.9.27/tests/tests2", nil},
-		{cfg, "benchmarksgame-team.pages.debian.net", nil},
+		{cfg, "benchmarksgame-team.pages.debian.net", blacklistDebian},
 	} {
 		t.Run(v.dir, func(t *testing.T) {
 			f, o, s, n := testParse(t, v.cfg, "assets/"+v.dir, v.blacklist)
@@ -1324,20 +1375,34 @@ func testTranslateBug(t *testing.T, dir string, blacklist map[string]struct{}) {
 }
 
 func TestTranslate(t *testing.T) {
+	if nogcc {
+		t.Skip()
+	}
+
 	cfg := defaultCfg()
 	cfg.SysIncludePaths = append(cfg.SysIncludePaths, "Include") // benchmarksgame
 	cfg.FS = cfs
 	blacklistCompCert := map[string]struct{}{}
 	blacklistMakarov := map[string]struct{}{}
 	blacklistGCC := map[string]struct{}{
-		// Assertions are deprecated, not supported.
-		"950919-1.c": {},
-		// Type size too big
-		"991014-1.c": {},
+		"950919-1.c": {}, // Assertions are deprecated, not supported.
+		"991014-1.c": {}, // Type size too big
+	}
+	blacklistDebian := map[string]struct{}{
+		"nbody-4.c": {}, // _Float16 not supported
+		"nbody-8.c": {}, // _Float16 not supported
 	}
 	switch fmt.Sprintf("%s/%s", runtime.GOOS, runtime.GOARCH) {
-	case "linux/s390x", "linux/loong64":
+	case "linux/mips64le":
 		blacklistCompCert["aes.c"] = struct{}{} // Unsupported endianness.
+	case "linux/s390x":
+		blacklistCompCert["aes.c"] = struct{}{} // Unsupported endianness.
+	case "linux/loong64":
+		blacklistCompCert["aes.c"] = struct{}{} // Unsupported endianness.
+		blacklistGCC["20101011-1.c"] = struct{}{}
+		blacklistGCC["fp-cmp-1.c"] = struct{}{}
+		blacklistGCC["fp-cmp-2.c"] = struct{}{}
+		blacklistGCC["fp-cmp-3.c"] = struct{}{}
 	case "windows/arm64":
 		blacklistGCC["pr49218.c"] = struct{}{}     //TODO
 		blacklistGCC["pr54471.c"] = struct{}{}     //TODO
@@ -1353,6 +1418,11 @@ func TestTranslate(t *testing.T) {
 		blacklistGCC["pr93213.c"] = struct{}{}     //TODO
 		blacklistGCC["pr98474.c"] = struct{}{}     //TODO
 		blacklistMakarov["setjmp2.c"] = struct{}{} //TODO
+	case "windows/386":
+		blacklistGCC["pr30704.c"] = struct{}{} // '__int128' is not supported on this target
+	case "freebsd/amd64":
+		blacklistDebian["mandelbrot-6.c"] = struct{}{}
+		blacklistDebian["spectral-norm-5.c"] = struct{}{}
 	}
 	var files, ok, skip, fails int32
 	for _, v := range []struct {
@@ -1369,7 +1439,7 @@ func TestTranslate(t *testing.T) {
 		{cfg, "github.com/vnmakarov", blacklistMakarov},
 		{cfg, "sqlite-amalgamation", nil},
 		{cfg, "tcc-0.9.27/tests/tests2", nil},
-		{cfg, "benchmarksgame-team.pages.debian.net", nil},
+		{cfg, "benchmarksgame-team.pages.debian.net", blacklistDebian},
 	} {
 		t.Run(v.dir, func(t *testing.T) {
 			f, o, s, n := testTranslate(t, v.cfg, "assets/"+v.dir, v.blacklist)
@@ -1495,7 +1565,7 @@ func testTranslate(t *testing.T, cfg *Config, dir string, blacklist map[string]s
 
 	p.wg.Wait()
 	for _, v := range fails {
-		t.Log(v)
+		t.Logf("FAIL: %v", v)
 	}
 	// fmt.Fprintf(os.Stderr, "%v: files %v, skip %v, ok %v, fails %v\n", dir, files, skip, ok, len(fails))
 	t.Logf("files %v, skip %v, ok %v, fails %v", files, skip, ok, len(fails))
@@ -1595,7 +1665,7 @@ func untar(dst string, r io.Reader, canOverwrite func(fn string, fi os.FileInfo)
 			}
 		case tar.TypeSymlink, tar.TypeXGlobalHeader:
 			// skip
-		case tar.TypeReg, tar.TypeRegA:
+		case tar.TypeReg:
 			dir := filepath.Dir(filepath.Join(dst, hdr.Name))
 			if _, err := os.Stat(dir); err != nil {
 				if !os.IsNotExist(err) {
@@ -1771,7 +1841,7 @@ func TestMake(t *testing.T) {
 	case "netbsd":
 		cfg.cflags = "-I/usr/pkg/include"
 	}
-	cfg.cflags = strings.TrimSpace(cfg.cflags + LongDouble64Flag(goos, goarch))
+	cfg.cflags = strings.TrimSpace(cfg.cflags)
 	for _, v := range []struct {
 		archive string
 		dir     string
@@ -1794,7 +1864,7 @@ func TestMake(t *testing.T) {
 		{"git.postgresql.org/git/postgresql.tar.gz", "postgresql", cfg, all},
 		{"www.lua.org/ftp/lua-5.4.4.tar.gz", "lua-5.4.4", cfg.noConfigure(), all},
 	} {
-		if !filter(v.filter) {
+		if !filter(v.filter) || re != nil && !re.MatchString(v.archive) {
 			continue
 		}
 
@@ -1809,15 +1879,9 @@ func TestMake(t *testing.T) {
 	t.Logf("TOTAL: files %v, skip %v, ok %v, fails %v", files, skip, ok, fails)
 }
 
-var (
-	goos   = runtime.GOOS
-	goarch = runtime.GOARCH
-	osarch = fmt.Sprintf("%s/%s", goos, goarch)
-)
-
 func filter(f []string) bool {
 	for _, v := range f {
-		if v == goos || v == goarch || v == osarch {
+		if v == goos || v == goarch || v == target {
 			return true
 		}
 	}
@@ -1880,11 +1944,11 @@ func testMake(t *testing.T, archive, dir string, mcfg *makeCfg) (files, ok, skip
 	}
 	switch goos {
 	case "darwin", "freebsd", "netbsd", "openbsd":
-		if _, err := shell("gmake"); err != nil {
+		if _, err := shell("gmake", makeJ); err != nil {
 			t.Skipf("SKIP: gmake: %v", err)
 		}
 	default:
-		if _, err := shell("make"); err != nil {
+		if _, err := shell("make", makeJ); err != nil {
 			t.Skipf("SKIP: make: %v", err)
 		}
 	}
@@ -1901,7 +1965,6 @@ func testMake(t *testing.T, archive, dir string, mcfg *makeCfg) (files, ok, skip
 
 		var a []string
 		if err := json.NewDecoder(bytes.NewReader(b)).Decode(&a); err != nil {
-			trc("", a)
 			t.Fatal(err)
 		}
 
@@ -2260,10 +2323,839 @@ func TestIssue149(t *testing.T) {
 		{Name: "test.h", Value: src},
 	}
 
-	ast, err := Parse(cfg, sources)
+	if _, err := Parse(cfg, sources); err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+}
+
+func TestConst0(t *testing.T) {
+	const src = "int i;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
 	if err != nil {
 		t.Fatalf("parse error: %v", err)
 	}
 
-	t.Logf("\n%s", ast.TranslationUnit)
+	d := ast.Scope.Nodes["i"][0].(*Declarator)
+	if g, e := d.IsConst(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsConst(), false; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestConst1(t *testing.T) {
+	const src = "const int i;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["i"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsConst(), true; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsConst(), true; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestConst2(t *testing.T) {
+	const src = "const int *p;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["p"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsConst(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsConst(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().(*PointerType).Elem().Attributes().IsConst(), true; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestConst3(t *testing.T) {
+	const src = "int * const p;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["p"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsConst(), true; g != e {
+		t.Error(g, e) //TODO
+	}
+
+	if g, e := d.Type().Attributes().IsConst(), true; g != e {
+		t.Error(g, e) //TODO
+	}
+
+	if g, e := d.Type().(*PointerType).Elem().Attributes().IsConst(), false; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestConst4(t *testing.T) {
+	const src = "const int * const p;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["p"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsConst(), true; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsConst(), true; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().(*PointerType).Elem().Attributes().IsConst(), true; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestConst5(t *testing.T) {
+	const src = "const int *p[];\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["p"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsConst(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsConst(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().(*ArrayType).Elem().Attributes().IsConst(), true; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestVolatile0(t *testing.T) {
+	const src = "int i;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["i"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsVolatile(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsVolatile(), false; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestVolatile1(t *testing.T) {
+	const src = "volatile int i;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["i"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsVolatile(), true; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsVolatile(), true; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestVolatile2(t *testing.T) {
+	const src = "volatile int *p;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["p"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsVolatile(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsVolatile(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().(*PointerType).Elem().Attributes().IsVolatile(), true; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestVolatile3(t *testing.T) {
+	const src = "int * volatile p;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["p"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsVolatile(), true; g != e {
+		t.Error(g, e) //TODO
+	}
+
+	if g, e := d.Type().Attributes().IsVolatile(), true; g != e {
+		t.Error(g, e) //TODO
+	}
+
+	if g, e := d.Type().(*PointerType).Elem().Attributes().IsVolatile(), false; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestVolatile4(t *testing.T) {
+	const src = "volatile int * volatile p;\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["p"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsVolatile(), true; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsVolatile(), true; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().(*PointerType).Elem().Attributes().IsVolatile(), true; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestVolatile5(t *testing.T) {
+	const src = "volatile int *p[];\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["p"][0].(*Declarator)
+	t.Log(d.Type())
+	if g, e := d.IsVolatile(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().Attributes().IsVolatile(), false; g != e {
+		t.Error(g, e)
+	}
+
+	if g, e := d.Type().(*ArrayType).Elem().Attributes().IsVolatile(), true; g != e {
+		t.Error(g, e)
+	}
+}
+
+func TestBindDeclarations(t *testing.T) {
+	const src = "int x __attribute__((addr(42)));"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	var found bool
+	for _, v := range ast.Scope.Nodes["x"] {
+		d, ok := v.(*Declarator)
+		if !ok {
+			continue
+		}
+		found = true
+		attrs := d.Type().Attributes()
+		t.Logf("%T: %v", v, d.Type().Attributes())
+
+		if !attrs.IsAttrSet("addr") {
+			t.Fatalf("attrs.IsAttrSet(\"addr\") failed")
+		}
+		attrval := attrs.AttrValue("addr")
+		if attrval == nil {
+			t.Fatalf("attrs.AttrValue(\"addr\") returned nil")
+		}
+		if len(attrval) != 1 {
+			t.Fatalf("attrs.AttrValue(\"addr\") returned slice of wrong length")
+		}
+		t.Log(attrval)
+		intval, ok := attrval[0].(Int64Value)
+		if !ok {
+			t.Fatalf("attrs.AttrValue(\"addr\") returned wrong type: %T", attrval[0])
+		}
+		if intval != 42 {
+			t.Fatalf("attrs.AttrValue(\"addr\") returned wrong value")
+		}
+	}
+	if !found {
+		t.Fatalf("Could not find node associated with x")
+	}
+}
+
+func TestPointerIntComparison(t *testing.T) {
+	const src = "void f(){int x = ((char *)1) == 1;}\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	_, err = Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+}
+
+func TestRandomError(t *testing.T) {
+	const src = "void f(){struct T{unsigned char x[1];};(struct T){0};}\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: `int __predefined_declarator;`}, // Just to test without wchar_t definition
+		{Name: "test.h", Value: src},
+	}
+
+	if _, err = Translate(cfg, sources); err == nil {
+		t.Fatalf("expected Translate error: %v", err)
+	}
+}
+
+func TestRandomError2(t *testing.T) {
+	const src = "void f(){struct T{unsigned char x[1];};(struct T){0};}\n"
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: `int __predefined_declarator;`}, // Just to test without wchar_t definition
+		{Name: "test.h", Value: src},
+	}
+
+	cfg.DefaultSizeT = ULongLong
+	cfg.DefaultWcharT = Int
+	if _, err = Translate(cfg, sources); err != nil {
+		t.Fatalf("unexpected Translate error: %v", err)
+	}
+}
+
+func TestAttrValue(t *testing.T) {
+	const src = `
+char a1 __attribute__((b1(100)));
+__attribute__((b2(200))) short a2;
+int *a3 __attribute__((b3(300)));
+__attribute__((b4(400))) long *a4;
+`
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+
+	ast, err := Translate(cfg, sources)
+	if err != nil {
+		t.Fatalf("parse error: %v", err)
+	}
+
+	d := ast.Scope.Nodes["a1"][0].(*Declarator)
+	a := d.Type().Attributes()
+	vs := a.AttrValue("b1")
+	if len(vs) != 1 || vs[0] != Int64Value(100) {
+		t.Fatal(vs)
+	}
+
+	d = ast.Scope.Nodes["a2"][0].(*Declarator)
+	a = d.Type().Attributes()
+	vs = a.AttrValue("b2")
+	if len(vs) != 1 || vs[0] != Int64Value(200) {
+		t.Fatal(vs)
+	}
+
+	d = ast.Scope.Nodes["a3"][0].(*Declarator)
+	a = d.Type().Attributes()
+	vs = a.AttrValue("b3")
+	if len(vs) != 1 || vs[0] != Int64Value(300) {
+		t.Fatal(vs)
+	}
+
+	a = d.Type().(*PointerType).Elem().Attributes()
+	vs = a.AttrValue("b3")
+	if len(vs) != 0 {
+		t.Fatal(vs)
+	}
+
+	d = ast.Scope.Nodes["a4"][0].(*Declarator)
+	a = d.Type().Attributes()
+	vs = a.AttrValue("b4")
+	if len(vs) != 0 {
+		t.Fatal(vs)
+	}
+
+	a = d.Type().(*PointerType).Elem().Attributes()
+	vs = a.AttrValue("b4")
+	if len(vs) != 1 || vs[0] != Int64Value(400) {
+		t.Fatal(vs)
+	}
+}
+
+func TestCanonicalizeStrlitTail(t *testing.T) {
+	for i, test := range []struct{ a, b string }{
+		{``, ``},
+		{`a`, `a`},
+		{`\a`, `\a`},
+		{`\-`, `\-`},
+		{`\x1`, `\x01`},
+		{`\x12`, `\x12`},
+		{`\x123`, `\x123`},
+		{`\0`, `\000`},
+		{`\01`, `\001`},
+		{`\0123`, `\0123`},
+	} {
+		if g, e := string(canonicalizeStrlitTail([]byte(test.a))), test.b; g != e {
+			t.Errorf("%v: %q -> %q, want %q", i, test.a, g, e)
+		}
+	}
+}
+
+func TestVaArgs1(t *testing.T) {
+	//	parser.go:265:shift: TRC bug.c:7:9: U+E030 "g"
+	//	parser.go:265:shift: TRC bug.c:7:9: U+0028 '(' "("
+	//	parser.go:265:shift: TRC bug.c:7:9: U+E054 "\"foo\""
+	//	parser.go:265:shift: TRC bug.c:7:9: U+002C ',' ","	<--- comma not deleted
+	//	parser.go:265:shift: TRC bug.c:7:9: U+0029 ')' ")"
+	//	parser.go:265:shift: TRC bug.c:7:17: U+003B ';' ";"
+	//	parser.go:265:shift: TRC bug.c:8:1: U+007D '}' "}"
+	//
+	//	bug.c:7:9: unexpected ')', expected unary expression
+	//
+	//	https://gcc.gnu.org/onlinedocs/gcc-4.9.4/cpp/Variadic-Macros.html
+	cfg := defaultCfg()
+	cfg.noPredefinedDeclarator = true
+	_, err := Parse(cfg, []Source{
+		{Name: "bug.c", Value: `
+#define m(a, ...) g(a, ##__VA_ARGS__)
+
+void g() {}
+
+void f() {
+        m("foo");
+}
+`},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+}
+
+func TestVaArgs2(t *testing.T) {
+	//	parser.go:265:shift: TRC bug.c:7:9: U+E030 "g"
+	//	parser.go:265:shift: TRC bug.c:7:9: U+0028 '(' "("
+	//	parser.go:265:shift: TRC bug.c:7:9: U+E054 "\"foo\""
+	//	parser.go:265:shift: TRC bug.c:7:9: U+002C ',' ",42"	<--- malformed token
+	//	parser.go:265:shift: TRC bug.c:7:9: U+0029 ')' ")"
+	//	parser.go:265:shift: TRC bug.c:7:21: U+003B ';' ";"
+	//	parser.go:265:shift: TRC bug.c:8:1: U+007D '}' "}"
+	//
+	//	bug.c:7:9: unexpected ')', expected unary expression
+	cfg := defaultCfg()
+	cfg.noPredefinedDeclarator = true
+	_, err := Parse(cfg, []Source{
+		{Name: "bug.c", Value: `
+#define m(a, ...) g(a, ##__VA_ARGS__)
+
+void g() {}
+
+void f() {
+        m("foo", 42);
+}
+`},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+}
+
+func TestSingleBytewchar(t *testing.T) {
+	const src = `
+#include <stddef.h>
+wchar_t *c = L"!";
+`
+	cfg, err := NewConfig(runtime.GOOS, runtime.GOARCH)
+	if err != nil {
+		t.Fatalf("failed to create new config: %v", err)
+	}
+	sources := []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<predefined2>", Value: `
+#define __WCHAR_MAX__ 0xff
+#define __WCHAR_MIN__ 0
+#define __WCHAR_WIDTH__ 8
+#define __WCHAR_TYPE__ unsigned char
+#define __SIZEOF_WCHAR_T__ 1
+`},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "test.h", Value: src},
+	}
+	var ast *AST
+	if ast, err = Translate(cfg, sources); err != nil {
+		t.Fatalf("unexpected Translate error: %v", err)
+	}
+	nodes := findAllNodes("InitDeclarator", ast.TranslationUnit)
+	for _, n := range nodes {
+		decl := n.(*InitDeclarator)
+		if decl.Declarator.Name() == "c" {
+			var out Node
+			d := 100
+			findNode("PrimaryExpression", decl, 0, &out, &d)
+			pe := out.(*PrimaryExpression)
+			if _, ok := pe.Value().(StringValue); !ok {
+				t.Fatalf("value should be of type StringValue, but was of type %T", pe.Value())
+			}
+			return
+		}
+	}
+	t.Fatalf("Could not find InitDeclarator for value 'c'")
+}
+
+// https://gitlab.com/cznic/ccgo/-/issues/34
+func TestIssue34(t *testing.T) {
+	cfg := defaultCfg()
+	_, err := Translate(cfg, []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "bug.c", Value: `
+
+void* f() {
+	return 0;
+}
+
+void* g() {
+	return &(((int*)(f()))[1]);
+} 
+
+`},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+}
+
+func TestIssue34b(t *testing.T) {
+	cfg := defaultCfg()
+	_, err := Translate(cfg, []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "bug.c", Value: `
+
+typedef unsigned long int __uint64_t;
+typedef __uint64_t uint64_t;
+typedef uint64_t ecs_id_t;
+typedef ecs_id_t ecs_entity_t;
+
+typedef enum ecs_primitive_kind_t {
+ EcsBool = 1,
+ EcsChar,
+ EcsByte,
+ EcsU8,
+ EcsU16,
+ EcsU32,
+ EcsU64,
+ EcsI8,
+ EcsI16,
+ EcsI32,
+ EcsI64,
+ EcsF32,
+ EcsF64,
+ EcsUPtr,
+ EcsIPtr,
+ EcsString,
+ EcsEntity,
+ EcsId,
+ EcsPrimitiveKindLast = EcsId
+} ecs_primitive_kind_t;
+
+typedef struct ecs_primitive_desc_t {
+ ecs_entity_t entity; 
+ ecs_primitive_kind_t kind;
+} ecs_primitive_desc_t;
+
+// ----------------------------------------------------------------------------
+
+typedef unsigned long int __uint64_t;
+typedef __uint64_t uint64_t;
+typedef uint64_t ecs_id_t;
+typedef ecs_id_t ecs_entity_t;
+
+typedef enum ecs_primitive_kind_t {
+ EcsBool = 1,
+ EcsChar,
+ EcsByte,
+ EcsU8,
+ EcsU16,
+ EcsU32,
+ EcsU64,
+ EcsI8,
+ EcsI16,
+ EcsI32,
+ EcsI64,
+ EcsF32,
+ EcsF64,
+ EcsUPtr,
+ EcsIPtr,
+ EcsString,
+ EcsEntity,
+ EcsId,
+ EcsPrimitiveKindLast = EcsId
+} ecs_primitive_kind_t;
+
+typedef struct ecs_primitive_desc_t {
+ ecs_entity_t entity; 
+ ecs_primitive_kind_t kind;
+} ecs_primitive_desc_t;
+
+`},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+}
+
+func TestIssue34c(t *testing.T) {
+	cfg := defaultCfg()
+	_, err := Translate(cfg, []Source{
+		{Name: "<predefined>", Value: cfg.Predefined},
+		{Name: "<builtin>", Value: Builtin},
+		{Name: "bug.c", Value: `
+
+typedef struct ecs_world_t ecs_world_t;
+typedef unsigned long int __uint64_t;
+typedef __uint64_t uint64_t;
+typedef uint64_t ecs_id_t;
+typedef ecs_id_t ecs_entity_t;
+typedef struct ecs_serializer_t {
+ 
+ int (*value)(
+ const struct ecs_serializer_t *ser, 
+ ecs_entity_t type, 
+ const void *value); 
+
+ 
+ int (*member)(
+ const struct ecs_serializer_t *ser, 
+ const char *member); 
+
+ const ecs_world_t *world;
+ void *ctx;
+} ecs_serializer_t;
+
+// ----------------------------------------------------------------------------
+
+typedef struct ecs_world_t ecs_world_t;
+typedef unsigned long int __uint64_t;
+typedef __uint64_t uint64_t;
+typedef uint64_t ecs_id_t;
+typedef ecs_id_t ecs_entity_t;
+typedef struct ecs_serializer_t {
+ 
+ int (*value)(
+ const struct ecs_serializer_t *ser, 
+ ecs_entity_t type, 
+ const void *value); 
+
+ 
+ int (*member)(
+ const struct ecs_serializer_t *ser, 
+ const char *member); 
+
+ const ecs_world_t *world;
+ void *ctx;
+} ecs_serializer_t;
+
+`},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
 }
